(*----------------------------------------------------------------------------*
 *  Direct3D sample for DirectX 8.x demostrating how to process X-Files data  *
 *  using D3DX8 functions. More precisely I try to 'SMOOTH' input mesh with   *
 *  minimal error introduction on vertex positions                            *
 *                                                                            *
 *  Copyright (c) 2002 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 11-Apr-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit MainUnit;

{$I DirectX.inc}

interface

uses
  Windows, Messages,  D3DX8, D3DApp, D3DFont, D3DFile, D3DUtil, DXUtil, DXFile,
  D3D8;

type
  PMyVertex = ^TMyVertex;
  TMyVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
  end;

const
  D3DFVF_VERTEX = D3DFVF_XYZ or D3DFVF_NORMAL;

type
  TMeshToRender = (mrNormal, mrReNormalized, mrWelded);

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  PMyD3DApplication = ^CMyD3DApplication;
  CMyD3DApplication = object(TD3DApplication)
  private
    m_pFont: CD3DFont;          // Font for drawing text
    m_pObject: CD3DMesh;        // This is our mesh but already loaded from disk X-File
    fEnvMap: IDirect3DTexture8; // Environment map texture

    fNormEpsilon, fWeldEpsilon: Single;
    pmGenNormals, pmWeldedNormals: ID3DXMesh;

    fNormalsHeight: Single;
    NormalsCount: array[mrNormal..mrWelded] of DWORD;
    NormalsVB: array[mrNormal..mrWelded] of IDirect3DVertexBuffer8;

    MeshMtrl, NormalsMtrl: TD3DMaterial8;

    fDisplayHelp, fShowStats,
    fShowNormals, fIsEnvMapped, fIsWireframe: Boolean;
  public
    function InitDeviceObjects: HResult; virtual;
    function RestoreDeviceObjects: HResult; virtual;
    function InvalidateDeviceObjects: HResult; virtual;
    function DeleteDeviceObjects: HResult; virtual;
    function Render: HResult; virtual;
    function FinalCleanup: HResult; virtual;
    function ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord; Format: TD3DFormat): HResult; virtual;

    // this is the main procedure in example
    function ProcessXFile(out pmGenNormals, pmWeldedNormals: ID3DXMesh): HRESULT;
    procedure SaveMeshesToDisk;
    procedure GenerateNormals(Mesh: ID3DXMesh; NormalsHeight: Single;
      out VB: IDirect3DVertexBuffer8; out iNormalsCount: DWORD);
  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; virtual;
  end;

implementation

uses Strings;


{$IFDEF DX81}
function D3DXWeldVertices(pMesh: ID3DXMesh; pEpsilons: PD3DXWeldEpsilons;
  rgdwAdjacencyIn, rgdwAdjacencyOut, pFaceRemap: PDWord;
  ppVertexRemap: PID3DXBuffer): HResult; stdcall; external d3dx8dll;
{$ELSE}
function D3DXWeldVertices(pMesh: ID3DXMesh; fEpsilon: Single;
  rgdwAdjacencyIn, rgdwAdjacencyOut, pFaceRemap: PDWord;
  ppbufVertexRemap: PID3DXBuffer): HResult; stdcall; external d3dx8dll;
{$ENDIF}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create(0);
  m_strWindowTitle      := 'X-File processing by Clootie';
  m_bUseDepthBuffer     := True;
  m_bShowCursorWhenFullscreen := True;

  m_dwCreationWidth:= 600;
  m_dwCreationHeight:= 320;

  fNormEpsilon:= 0.07;
  fWeldEpsilon:= 0.1;
  fNormalsHeight:= 0.05;

  fDisplayHelp:= False;
  fShowStats:= True;
  fShowNormals:= False;
  fIsEnvMapped:= False;
  fIsWireframe:= False;

  New(m_pFont, Create('Arial', 12, D3DFONT_BOLD));
  New(m_pObject, CreateNoName);
  m_pObject^.m_bUseMaterials:= False;

  // Setup materials
  D3DUtil_InitMaterial(MeshMtrl, 1.0, 1.0, 1.0, 1.0);
  D3DUtil_InitMaterial(NormalsMtrl, 1.0, 1.0, 1.0, 1.0);
  NormalsMtrl.Emissive:= D3DXColor(0, 0.9, 0.9, 0);
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render;
{nested} procedure RenderMesh(Style: TMeshToRender);
begin
  m_pd3dDevice.SetMaterial(MeshMtrl);
  case Style of
    mrNormal:
      m_pObject^.Render(m_pd3dDevice, True, True);
    mrReNormalized:
      if pmGenNormals <> nil then pmGenNormals.DrawSubset(0);
    mrWelded:
      if pmWeldedNormals <> nil then pmWeldedNormals.DrawSubset(0);
  end;

  if fShowNormals then
  begin
    m_pd3dDevice.SetMaterial(NormalsMtrl);

    m_pd3dDevice.SetVertexShader(D3DFVF_XYZ);
    m_pd3dDevice.SetStreamSource(0, NormalsVB[Style], SizeOf(TD3DXVector3));
    m_pd3dDevice.DrawPrimitive(D3DPT_LINELIST, 0, NormalsCount[Style]);

    m_pd3dDevice.SetMaterial(MeshMtrl);
  end;
end;
{nested} procedure RenderTwoFace(Style: TMeshToRender);
var
  m: TD3DXMatrix;
begin
  if fIsEnvMapped then
  begin
    m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iFalse);
    m_pd3dDevice.SetTexture(0, fEnvMap);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACENORMAL);

    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
  end else
  begin
    m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);
    m_pd3dDevice.SetTexture(0, nil);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU);

    m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG2);
    m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG2);
  end;

  D3DXMatrixRotationY(m, -m_fTime);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m);

  m_pd3dDevice.MultiplyTransform(D3DTS_WORLD,
    D3DXMatrixRotationX(m, -D3DX_PI*0.5)^);

  m_pd3dDevice.MultiplyTransform(D3DTS_WORLD,
    D3DXMatrixTranslation(m, 0, -0.2, 0)^);

  // Render first "human face"
  RenderMesh(Style);

  m_pd3dDevice.MultiplyTransform(D3DTS_WORLD,
    D3DXMatrixTranslation(m, 0, 0.4, 0)^);
  m_pd3dDevice.MultiplyTransform(D3DTS_WORLD,
    D3DXMatrixRotationZ(m, D3DX_PI)^);

  // Render second "human face"
  RenderMesh(Style);
end;

var
  m: TD3DXMatrix;
var
  S: String; //[80];
begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
    $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

    D3DXMatrixTranslation(m, -0.9, 0, 2);
    m_pd3dDevice.SetTransform(D3DTS_VIEW, m);
    RenderTwoFace(mrNormal);

    D3DXMatrixTranslation(m,  0.0, 0, 2);
    m_pd3dDevice.SetTransform(D3DTS_VIEW, m);
    RenderTwoFace(mrReNormalized);

    D3DXMatrixTranslation(m,  0.9, 0, 2);
    m_pd3dDevice.SetTransform(D3DTS_VIEW, m);
    RenderTwoFace(mrWelded);

    // Output statistics
    m_pFont^.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats, 0);

    if fDisplayHelp then
    begin
      m_pFont^.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0),
        'Press ''H'' to hide help', 0);

      m_pFont^.DrawText(7, 40, D3DCOLOR_ARGB(255,255,255,255),
        'A / Z'#10 +
        'S / X'#10 +
        'C'#10 +
        'E'#10 +
        'N'#10 +
        'D'#10 +
        'Space'#10 +
        'Enter', 0);

      m_pFont^.DrawText(80, 40, D3DCOLOR_ARGB(255,255,255,0),
        'Modify epsilon for D3DXComputeNormals()'#10 +
        'Modify epsilon for D3DXWeldVertices()'#10 +
        'Show / hide statistics'#10 +
        'Lighting / Environment mapping'#10 +
        'Show / hide mesh normals'#10 +
        'Save processed meshed to disk'#10 +
        'Pause / Still'#10 +
        'Run', 0);
    end else
    begin
      if fShowStats then
      begin
        m_pFont^.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0),
          'Press ''H'' to display help'#10 +
          'Epsilons:'#10 +
          #10#10 +
          'Faces/Vert:', 0);

        S:= Fix(fNormEpsilon, 0, 4) + #10 + Fix(fWeldEpsilon, 0, 4) + #10#10 +
          IntToStr(m_pObject^.GetSysMemMesh.GetNumFaces) + ' / ' +
            IntToStr(m_pObject^.GetSysMemMesh.GetNumVertices) + #10 +
          IntToStr(pmGenNormals.GetNumFaces) + ' / ' +
            IntToStr(pmGenNormals.GetNumVertices) + #10 +
          IntToStr(pmWeldedNormals.GetNumFaces) + ' / ' +
            IntToStr(pmWeldedNormals.GetNumVertices) + #0;

        m_pFont^.DrawText(95, 40, D3DCOLOR_ARGB(255,255,255,255), PChar(@S[1]), 0);

        m_pFont^.DrawText(180, 40, D3DCOLOR_ARGB(255,255,255,0),
          'for D3DXComputeNormals()'#10 +
          'for D3DXWeldVertices()'#10 +
          #10 +
          'in original mesh'#10 +
          'after D3DXComputeNormals()'#10 +
          'after D3DXWeldVertices()'#10, 0);
      end else
        m_pFont^.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0),
          'Press ''H'' to display help', 0);
    end;

    m_pFont^.DrawText(80, m_rcWindowClient.Bottom - 50,
      D3DCOLOR_ARGB(255,255,100,255), 'Initial mesh', 0);
    m_pFont^.DrawText(m_rcWindowClient.Right div 2 - 80, m_rcWindowClient.Bottom - 60,
      D3DCOLOR_ARGB(255,255,155,255), 'After'#10'D3DXComputeNormals()', 0);
    m_pFont^.DrawText(m_rcWindowClient.Right - 170, m_rcWindowClient.Bottom - 60,
      D3DCOLOR_ARGB(255,255,200,255), 'After'#10'D3DXWeldVertices()', 0);

    m_pFont^.DrawText(2, m_rcWindowClient.Bottom - 20,
      D3DCOLOR_ARGB(255,255,255,255),
      'Visit  "http://clootie.narod.ru"  for more demos...', 0);

    // End the scene
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects;
begin
  // Restore the font
  m_pFont^.InitDeviceObjects(m_pd3dDevice);

  D3DXCreateTextureFromFile(m_pd3dDevice, 'Chrome.bmp', fEnvMap);

  // Load mesh genearated by us and show it...
  if FAILED(m_pObject^.Create_(m_pd3dDevice, PChar('Face.x'))) then
  begin
    Result:= D3DAPPERR_MEDIANOTFOUND;
    Exit;
  end;
  m_pObject^.SetFVF(m_pd3dDevice, D3DFVF_VERTEX);

  GenerateNormals(m_pObject^.GetSysMemMesh, fNormalsHeight,
    NormalsVB[mrNormal], NormalsCount[mrNormal]);

  // Let's do it
  ProcessXFile(pmGenNormals, pmWeldedNormals);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects;
function D3DXColorFromDWord(c: DWord): TD3DXColor;
const
  f: Single = 1/255;
begin
  with Result do
  begin
    r:= f * Byte(c shr 16);
    g:= f * Byte(c shr  8);
    b:= f * Byte(c shr  0);
    a:= f * Byte(c shr 24);
  end;
end;
var
  matProj, mat: TD3DXMatrix;
  fAspect: Single;
  light: TD3DLight8;
begin
  m_pFont^.RestoreDeviceObjects;

  // Restore device-memory objects for the mesh
  m_pObject^.RestoreDeviceObjects(m_pd3dDevice);

  // Set up the textures
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG2);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG2);

  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

  // Set miscellaneous render states
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,        iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,        $00444444);

  // Generate spheremap texture coords, and shift them over
  mat._11 := 0.5; mat._12 := 0.0; mat._13 := 0.0; mat._14 := 0.0;
  mat._21 := 0.0; mat._22 :=-0.5; mat._23 := 0.0; mat._24 := 0.0;
  mat._31 := 0.0; mat._32 := 0.0; mat._33 := 1.0; mat._34 := 0.0;
  mat._41 := 0.5; mat._42 := 0.5; mat._43 := 0.0; mat._44 := 1.0;

  m_pd3dDevice.SetTransform(D3DTS_TEXTURE0, mat);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2);

  // Set the world matrix
  D3DXMatrixIdentity(mat);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, mat);

  // Set the projection matrix
  fAspect:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Set up lighting states
  D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, -1.0, 1.0, 1.0);
  light.Diffuse:= D3DXColorFromDWord($00928489);

  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, True);
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);

  // Set mesh material
  m_pd3dDevice.SetMaterial(MeshMtrl);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects;
begin
  m_pFont^.InvalidateDeviceObjects;
  m_pObject^.InvalidateDeviceObjects;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects;
begin
  SAFE_RELEASE(pmGenNormals);
  SAFE_RELEASE(pmWeldedNormals);

  m_pFont^.DeleteDeviceObjects;
  m_pObject^.Destroy_;

  SAFE_RELEASE(fEnvMap);
  
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pObject);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord;
  Format: TD3DFormat): HResult;
begin
  if dwBehavior and D3DCREATE_PUREDEVICE <> 0 then
  begin
    Result:= E_FAIL; // GetTransform doesn't work on PUREDEVICE
    Exit;
  end;

  // If this is a TnL device, make sure it supports directional lights
  if (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0) or
     (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0) then
  begin
    if (pCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS = 0) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
begin
  if (WM_KEYDOWN = uMsg) then
  begin
    if (Ord('H') = wParam) then fDisplayHelp := not fDisplayHelp;
    if (Ord('C') = wParam) then fShowStats   := not fShowStats;
    if (Ord('E') = wParam) then fIsEnvMapped := not fIsEnvMapped;
    if (Ord('N') = wParam) then fShowNormals := not fShowNormals;
    if (Ord('D') = wParam) then SaveMeshesToDisk;


    if (Ord('A') = wParam) then fNormEpsilon:= fNormEpsilon * 1.1;
    if (Ord('Z') = wParam) then fNormEpsilon:= fNormEpsilon * 1/1.1;
    if Char(wParam) in ['A', 'Z'] then
    begin // Let's do it again!!!
      ProcessXFile(pmGenNormals, pmWeldedNormals);
    end;

    if (Ord('S') = wParam) then fWeldEpsilon:= fWeldEpsilon * 1.1;
    if (Ord('X') = wParam) then fWeldEpsilon:= fWeldEpsilon * 1/1.1;
    if Char(wParam) in ['S', 'X'] then
    begin // Let's do it again!!!
      ProcessXFile(pmGenNormals, pmWeldedNormals);
    end;

    if (Ord('W') = wParam) then // Wireframe mode
    begin
      fIsWireframe:= not fIsWireframe;
      if fIsWireframe then
        m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME)
      else
        m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

function CMyD3DApplication.ProcessXFile(out pmGenNormals, pmWeldedNormals: ID3DXMesh): HRESULT;
var
  pAdjacencyBuffer: ID3DXBuffer;
  pmGenNormalsSys, pmWeldedNormalsSys: ID3DXMesh;

  MeshDecl: TFVFDeclaration;
{$IFDEF DX81}
  WeldEpsilons: TD3DXWeldEpsilons;
{$ENDIF}
begin
  pmGenNormals:= nil;
  pmWeldedNormals:= nil;
  pAdjacencyBuffer:= nil;

  Result:= D3DXLoadMeshFromX('Face.x', D3DXMESH_SYSTEMMEM, m_pd3dDevice, nil,
                             nil, nil, pmGenNormalsSys);
  if FAILED(Result) then Exit;

  // Let's try to optimize our model
  // To do it we need to have adjacency information about mesh, so generate it
  D3DXCreateBuffer(3 * pmGenNormalsSys.GetNumFaces * SizeOf(DWORD), pAdjacencyBuffer);
  pmGenNormalsSys.GenerateAdjacency(fNormEpsilon, pAdjacencyBuffer.GetBufferPointer);

  D3DXComputeNormals(pmGenNormalsSys{$IFDEF DX81}, pAdjacencyBuffer.GetBufferPointer{$ENDIF});

  pmGenNormalsSys.OptimizeInplace(
    D3DXMESHOPT_COMPACT or D3DXMESHOPT_ATTRSORT or D3DXMESHOPT_VERTEXCACHE,
    pAdjacencyBuffer.GetBufferPointer, nil, nil, nil);

  pmGenNormalsSys.GetDeclaration(MeshDecl);

  if FAILED(pmGenNormalsSys.CloneMesh(D3DXMESH_MANAGED, @MeshDecl,
    m_pd3dDevice, pmGenNormals))
  then Exit;

  GenerateNormals(pmGenNormals, fNormalsHeight,
    NormalsVB[mrReNormalized], NormalsCount[mrReNormalized]);
//////////////////////////////////////////////////////////////////////////////

  Result:= D3DXLoadMeshFromX('Face.x', D3DXMESH_SYSTEMMEM, m_pd3dDevice, nil,
                             nil, nil, pmWeldedNormalsSys);
  if FAILED(Result) then Exit;

{$IFDEF DX81}
  FillMemory(@WeldEpsilons, SizeOf(WeldEpsilons), 0);
  WeldEpsilons.Flags:= D3DXWELDEPSILONS_WELDPARTIALMATCHES;
  WeldEpsilons.Normal:= fWeldEpsilon;
  D3DXWeldVertices(pmWeldedNormalsSys, @WeldEpsilons, nil, nil, nil, nil);
{$ELSE}
  D3DXWeldVertices(pmWeldedNormalsSys, fWeldEpsilon, nil, nil, nil, nil);
{$ENDIF}

  pmWeldedNormalsSys.OptimizeInplace(
    D3DXMESHOPT_COMPACT or D3DXMESHOPT_ATTRSORT or D3DXMESHOPT_VERTEXCACHE,
    pAdjacencyBuffer.GetBufferPointer, nil, nil, nil);

  pmWeldedNormalsSys.GetDeclaration(MeshDecl);
  if FAILED(pmWeldedNormalsSys.CloneMesh(D3DXMESH_MANAGED, @MeshDecl,
    m_pd3dDevice, pmWeldedNormals))
  then Exit;

  GenerateNormals(pmWeldedNormals, fNormalsHeight,
    NormalsVB[mrWelded], NormalsCount[mrWelded]);

  SAFE_RELEASE(pmGenNormalsSys);
  SAFE_RELEASE(pmWeldedNormalsSys);
  SAFE_RELEASE(pAdjacencyBuffer);
end;

// Create X-File in text format - they are easy to read
procedure CMyD3DApplication.SaveMeshesToDisk;
begin
  // for mesh after D3DXComputeNormals
  D3DXSaveMeshToX('Face2Renormalize.x', pmGenNormals, nil, nil, 0,
    DWORD(DXFILEFORMAT_TEXT));
  // for mesh after D3DXWeldVertices
  D3DXSaveMeshToX('Face2Weld.x', pmWeldedNormals, nil, nil, 0,
    DWORD(DXFILEFORMAT_TEXT));
end;

// This procedure returns Count of Normals in Mesh and
// VertexBuffer containing Line List of normals
procedure CMyD3DApplication.GenerateNormals(Mesh: ID3DXMesh;
  NormalsHeight: Single; out VB: IDirect3DVertexBuffer8; out iNormalsCount: DWORD);
var
  i: Integer;
  dwFVF, VStride: DWORD;
  NormalShift: Integer;
  VData: PByte;
  NData: PD3DXVector3;
  t: TD3DXVector3;
begin
  // *** NOTE *** (Clootie)
  // Routine internally assumes what every vertex is used during rendering
  // (no "hanging" vertices)

  dwFVF:= Mesh.GetFVF;

  NormalShift:= 0;
  if dwFVF and D3DFVF_XYZ   = D3DFVF_XYZ   then NormalShift:= 3;
  if dwFVF and D3DFVF_XYZB1 = D3DFVF_XYZB1 then NormalShift:= 4;
  if dwFVF and D3DFVF_XYZB2 = D3DFVF_XYZB2 then NormalShift:= 5;
  if dwFVF and D3DFVF_XYZB3 = D3DFVF_XYZB3 then NormalShift:= 6;
  if dwFVF and D3DFVF_XYZB4 = D3DFVF_XYZB4 then NormalShift:= 7;
  if dwFVF and D3DFVF_XYZB5 = D3DFVF_XYZB5 then NormalShift:= 8;

  VStride:= D3DXGetFVFVertexSize(dwFVF);
  iNormalsCount:= Mesh.GetNumVertices;


  if FAILED(m_pd3dDevice.CreateVertexBuffer(SizeOf(TD3DXVector3)*2*iNormalsCount,
    D3DUSAGE_WRITEONLY, D3DFVF_XYZ, D3DPOOL_MANAGED, VB))
  then Exit;

  if FAILED(vb.Lock(0, 0, PByte(NData), D3DLOCK_NOSYSLOCK))
  then Exit;

  Mesh.LockVertexBuffer(D3DLOCK_NOSYSLOCK or D3DLOCK_READONLY, VData);
  {}
    for i:= 0 to iNormalsCount - 1 do
    begin
      NData^:= PD3DXVector3(VData)^;
      Inc(NData);

      D3DXVec3Add(NData^,
        PD3DXVector3(VData)^,
        D3DXVec3Scale(t,
          PD3DXVector3(PByte(Integer(VData) + NormalShift*SizeOf(Single)))^,
          NormalsHeight)^);
      Inc(NData);

      Inc(VData, VStride);
    end;
  {}

    Mesh.UnlockVertexBuffer;
    vb.Unlock;
  {}
end;

end.
